#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "flint/flint.h"
#include "flint/fmpz.h"
#include "flint/fmpz_mod.h"
#include "flint/fmpz_mod_poly.h"
#include "flint/fq.h"
#include "flint/fq_mat.h"
#include "flint/fq_poly.h"
#include "flint/ulong_extras.h"

typedef enum {
    Peterson_Gorenstein_Zierler,
    Berlekamp_Massey,
    Sugiyama,
} DecodingAlgorithm;

fq_ctx_t ctx;
fmpz_t p;
fq_t alpha, one, result, *alpha_pows;
fq_poly_t g, c, sigma, e, c_hat, r, f_hat, *y, *f;
slong n, m, t, d, k, message[256], error_locations[256], num_errors,
    num_messages = 1, print = 0;

void init() {
    // Inicialitzar el context de cos finit per FLINT.
    fmpz_init_set_ui(p, 2);
    fq_ctx_init(ctx, p, m, "X");
    fq_init(one, ctx);
    fq_one(one, ctx);

    n = n_pow(2, m) - 1;

    // Precomputar les potencies de alpha.
    fq_init(alpha, ctx);
    fq_gen(alpha, ctx);
    alpha_pows = flint_malloc(n * sizeof(fq_t));
    for (slong i = 0; i < n; i++) {
        fq_init(alpha_pows[i], ctx);
        fq_pow_ui(alpha_pows[i], alpha, i, ctx);
    }

    // Generar les classes 2-ciclotomiques.
    slong cosets[64][8] = {0}, size[64] = {0}, gen_cosets[64] = {0},
          powers[64] = {0};
    slong i = 1, j, temp, matched, next_repr, number_cosets, redundancy = 0,
          number_gen_cosets;
    cosets[1][0] = 1;
    size[0] = 1;
    size[1] = 1;
    do {
        // Generar la classe C_i.
        j = 0;
        temp = cosets[i][j] * 2 % n;
        while (temp != cosets[i][0]) {
            cosets[i][++j] = temp;
            temp = temp * 2 % n;
            size[i]++;
        }
        // Trobar el seguent representant.
        next_repr = 2;
        do {
            next_repr++;
            matched = 0;
            for (slong ii = 1; ii <= i && !matched; ii++) {
                for (slong jj = 0; (jj < size[ii]) && !matched; jj++) {
                    if (cosets[ii][jj] == next_repr) {
                        matched = 1;
                    }
                }
            }
        } while (matched && next_repr < n);
        // Si s'ha trobat un representant, inicialitzar com el primer element.
        if (!matched) {
            cosets[++i][0] = next_repr;
            size[i] = 1;
        }
        // Repetir fins que s'hagin trobat totes les classes.
    } while (next_repr < n);
    number_cosets = i;

    d = 2 * t + 1;

    // Trobar quines classes conte els valors 1, 2, ..., d-1.
    temp = 0;
    for (i = 1; i <= number_cosets; i++) {
        matched = 0;
        for (j = 0; (j < size[i]) && !matched; j++) {
            if (cosets[i][j] < d) {
                matched = 1;
                gen_cosets[temp++] = i;
            }
        }
        // Si aquesta classe conte algun valor, ajustar el nombre de
        // redundancia.
        if (matched) {
            redundancy += size[i];
        }
    }
    number_gen_cosets = temp;
    k = n - redundancy;

    if (k < 1) {
        flint_printf("No es pot construir el codi.\n");
        exit(1);
    }

    if (print) flint_printf("Codi BCH binari [%wd, %wd, %wd]\n", n, k, d);

    // Obtenir els elements de les classes trobades.
    temp = 0;
    for (i = 0; i < number_gen_cosets; i++) {
        for (j = 0; j < size[gen_cosets[i]]; j++) {
            powers[temp++] = cosets[gen_cosets[i]][j];
        }
    }

    // Obtenir el polinomi generador del codi.
    fq_poly_init(g, ctx);
    fq_poly_one(g, ctx);
    fq_poly_t h;
    fq_poly_init(h, ctx);
    for (i = 0; i < redundancy; i++) {
        fq_poly_gen(h, ctx);
        fq_poly_set_coeff(h, 0, alpha_pows[powers[i]], ctx);
        fq_poly_mul(g, g, h, ctx);
    }
    fq_poly_clear(h, ctx);
    if (print) {
        flint_printf("Polinomi generator:\t");
        fq_poly_print_pretty(g, "X", ctx);
        flint_printf("\n");
    }

    f = flint_malloc(num_messages * sizeof(fq_poly_t));
    y = flint_malloc(num_messages * sizeof(fq_poly_t));
    for (i = 0; i < num_messages; i++) {
        fq_poly_init(f[i], ctx);
        fq_poly_init(y[i], ctx);
    }
    fq_poly_init(c, ctx);
    fq_poly_init(sigma, ctx);
    fq_poly_init(e, ctx);
    fq_poly_init(c_hat, ctx);
    fq_poly_init(r, ctx);
    fq_poly_init(f_hat, ctx);
    fq_init(result, ctx);
}

void clear() {
    // Netegar la memòria.
    fq_poly_clear(c, ctx);
    fq_poly_clear(g, ctx);
    fq_poly_clear(sigma, ctx);
    fq_poly_clear(e, ctx);
    fq_poly_clear(c_hat, ctx);
    fq_poly_clear(r, ctx);
    fq_poly_clear(f_hat, ctx);
    for (slong i = 0; i < n; i++) {
        fq_clear(alpha_pows[i], ctx);
    }
    for (slong i = 0; i < num_messages; i++) {
        fq_poly_clear(f[i], ctx);
        fq_poly_clear(y[i], ctx);
    }
    flint_free(alpha_pows);
    flint_free(f);
    flint_free(y);
    fq_clear(one, ctx);
    fq_clear(alpha, ctx);
    fq_clear(result, ctx);
    fq_ctx_clear(ctx);
    fmpz_clear(p);
}

void encode(slong index) {
    fq_poly_zero(f[index], ctx);
    // Obtenir el missatge en forma polinomial, f(x).
    for (slong i = 0; i < k; i++) {
        if (message[i]) {
            fq_poly_set_coeff(f[index], i, one, ctx);
        }
    }
    if (print) {
        flint_printf("Missatge:\t\t");
        fq_poly_print_pretty(f[index], "X", ctx);
        flint_printf("\n");
    }

    // Obtenir hf(x) = x^{n-k} * f(x).
    fq_poly_t h;
    fq_poly_init(h, ctx);
    fq_poly_set_coeff(h, n - k, one, ctx);
    fq_poly_t hf;
    fq_poly_init(hf, ctx);
    fq_poly_mul(hf, f[index], h, ctx);

    // Obtenir el residu de dividir hf(x) entre g(x).
    fq_poly_t r;
    fq_poly_init(r, ctx);
    fq_poly_rem(r, hf, g, ctx);

    // Obtenir el missatge codificat c(x) = hf(x) - r(x).
    fq_poly_sub(c, hf, r, ctx);

    if (print) {
        flint_printf("Codi:\t\t\t");
        fq_poly_print_pretty(c, "X", ctx);
        flint_printf("\n");
    }

    // Netegar la memoria.
    fq_poly_clear(h, ctx);
    fq_poly_clear(hf, ctx);
    fq_poly_clear(r, ctx);
}

void introduce_error(slong index) {
    fq_poly_t e;
    fq_poly_init(e, ctx);
    // Obtenir el polinomi d'error, e(x).
    for (slong i = 0; i < num_errors; i++) {
        fq_poly_set_coeff(e, error_locations[i], one, ctx);
    }

    if (print) {
        flint_printf("Error:\t\t\t");
        fq_poly_print_pretty(e, "X", ctx);
        flint_printf("\n");
    }

    // Obtenir el vector rebut, y(x) = c(x) + e(x).
    fq_poly_zero(y[index], ctx);
    fq_poly_add(y[index], c, e, ctx);

    if (print) {
        flint_printf("Vector rebut:\t\t");
        fq_poly_print_pretty(y[index], "X", ctx);
        flint_printf("\n");
    }

    // Netegar la memòria.
    fq_poly_clear(e, ctx);
}

slong decode_PGZ(fq_t syndromes[]) {
    // Obtenir la matriu M.
    fq_mat_t M;
    fq_mat_init(M, t, t, ctx);
    for (slong i = 0; i < t; i++) {
        for (slong j = 0; j < t; j++) {
            fq_mat_entry_set(M, i, j, syndromes[i + j + 1], ctx);
        }
    }
    // Decrementar mu fins que la matriu sigui invertible.
    slong mu = t + 1, invertible = 0;
    fq_mat_t M_inv, M_mu;
    while (!invertible && mu > 1) {
        mu--;
        fq_mat_window_init(M_mu, M, 0, 0, mu, mu, ctx);
        fq_mat_init(M_inv, mu, mu, ctx);
        invertible = fq_mat_inv(M_inv, M_mu, ctx);
        // Netegar la memoria.
        if (!invertible) {
            fq_mat_clear(M_inv, ctx);
            fq_mat_window_clear(M_mu, ctx);
        }
    }
    if (!invertible) {
        // Netegar la memoria.
        fq_mat_window_clear(M_mu, ctx);
        fq_mat_clear(M_inv, ctx);
        fq_mat_clear(M, ctx);
        if (print)
            flint_printf(
                "No s'ha pogut corregir l'error, ha fallat al pas 2.\n");
        return 1;
    }
    slong nu = mu;
    // Resoldre l'equacio matricial per obtenir sigma, el polinomi localitzador
    // d'error.
    fq_mat_t V, S;
    fq_mat_init(V, nu, 1, ctx);
    fq_mat_init(S, nu, 1, ctx);
    for (slong i = 0; i < nu; i++) {
        fq_mat_entry_set(V, i, 0, syndromes[nu + i + 1], ctx);
    }
    fq_mat_mul(S, M_inv, V, ctx);
    fq_poly_set_coeff(sigma, 0, one, ctx);
    for (slong i = 0; i < nu; i++) {
        fq_poly_set_coeff(sigma, i + 1, fq_mat_entry(S, nu - i - 1, 0), ctx);
    }

    if (print) {
        flint_printf("Sigma:\t\t\t");
        fq_poly_print_pretty(sigma, "X", ctx);
        flint_printf("\n");
    }

    // Netegar la memoria.
    fq_mat_clear(S, ctx);
    fq_mat_clear(V, ctx);
    fq_mat_window_clear(M_mu, ctx);
    fq_mat_clear(M_inv, ctx);
    fq_mat_clear(M, ctx);
    return 0;
}

slong decode_BM(fq_t syndromes[]) {
    // Inicialitzar els polinomis sigma i les discrepancies.
    fq_poly_t sigma_mu, sigma_mu_next, sigma_rho;
    fq_t discrepancy_mu, discrepancy_mu_next, discrepancy_rho, temp;
    slong exponent, degree_mu, rho = 0, rho_value = -1;
    fq_poly_init(sigma_mu, ctx);
    fq_poly_init(sigma_mu_next, ctx);
    fq_poly_init(sigma_rho, ctx);
    fq_init(discrepancy_mu, ctx);
    fq_init(discrepancy_mu_next, ctx);
    fq_init(discrepancy_rho, ctx);
    fq_init(temp, ctx);
    fq_poly_one(sigma_mu, ctx);
    fq_poly_one(sigma_rho, ctx);
    fq_set(discrepancy_mu, syndromes[1], ctx);
    fq_one(discrepancy_rho, ctx);

    for (slong mu = 1; mu < t + 1; mu++) {
        // Pas iteratiu.
        // Obtenir el polinomi sigma^{(mu+1)}.
        if (fq_is_zero(discrepancy_mu, ctx)) {
            fq_poly_set(sigma_mu_next, sigma_mu, ctx);
        } else {
            fq_inv(temp, discrepancy_rho, ctx);
            fq_mul(temp, discrepancy_mu, temp, ctx);
            if (rho == 0) {
                exponent = 2 * (mu - 1) + 1;
            } else {
                exponent = 2 * (mu - 1 - (rho - 1));
            }
            fq_poly_set_coeff(sigma_mu_next, exponent, temp, ctx);
            fq_poly_mul(sigma_mu_next, sigma_mu_next, sigma_rho, ctx);
            fq_poly_add(sigma_mu_next, sigma_mu, sigma_mu_next, ctx);
        }
        // Obtenir la discrepancia Delta_{mu+1}.
        if (mu == t) break;
        for (slong j = 1; j < 2 * mu; j++) {
            fq_poly_get_coeff(temp, sigma_mu_next, j, ctx);
            fq_mul(temp, temp, syndromes[2 * mu + 1 - j], ctx);
            fq_add(discrepancy_mu_next, discrepancy_mu_next, temp, ctx);
        }
        fq_poly_get_coeff(temp, sigma_mu, 2 * mu + 1, ctx);
        fq_add(discrepancy_mu_next, discrepancy_mu_next, temp, ctx);
        fq_add(discrepancy_mu_next, discrepancy_mu_next, syndromes[2 * mu + 1],
               ctx);
        // Actualitzar rho.
        degree_mu = fq_poly_degree(sigma_mu, ctx);
        if (!fq_is_zero(discrepancy_mu, ctx) &&
            2 * (mu - 1) - degree_mu > rho_value) {
            rho = mu;
            rho_value = 2 * (mu - 1) - degree_mu;
            fq_poly_set(sigma_rho, sigma_mu, ctx);
            fq_set(discrepancy_rho, discrepancy_mu, ctx);
        }
        // Intercanviar els valors per la seguent iteracio.
        fq_poly_set(sigma_mu, sigma_mu_next, ctx);
        fq_poly_zero(sigma_mu_next, ctx);
        fq_set(discrepancy_mu, discrepancy_mu_next, ctx);
        fq_zero(discrepancy_mu_next, ctx);
    }

    // Obtenir sigma.
    fq_poly_set(sigma, sigma_mu_next, ctx);
    // Netejar la memoria.
    fq_clear(temp, ctx);
    fq_clear(discrepancy_rho, ctx);
    fq_clear(discrepancy_mu_next, ctx);
    fq_clear(discrepancy_mu, ctx);
    fq_poly_clear(sigma_rho, ctx);
    fq_poly_clear(sigma_mu_next, ctx);
    fq_poly_clear(sigma_mu, ctx);

    if (fq_poly_degree(sigma, ctx) > t) {
        if (print)
            flint_printf(
                "No s'ha pogut corregir l'error, ha fallat al pas 2.\n");
        return 1;
    }
    if (print) {
        flint_printf("Sigma:\t\t\t");
        fq_poly_print_pretty(sigma, "X", ctx);
        flint_printf("\n");
    }

    return 0;
}

slong decode_Sugiyama(fq_t syndromes[]) {
    // Inicialitzar r_{-1}(x), r_0(x), b_{-1}(x), b_0(x).
    fq_poly_t r, r_prev, b, b_prev, h;
    fq_poly_init(r, ctx);
    fq_poly_init(r_prev, ctx);
    fq_poly_init(b, ctx);
    fq_poly_init(b_prev, ctx);
    fq_poly_init(h, ctx);
    fq_poly_set_coeff(r_prev, 2 * t, one, ctx);
    for (slong i = 0; i < 2 * t; i++) {
        fq_poly_set_coeff(r, i, syndromes[i + 1], ctx);
    }
    fq_poly_set_coeff(b, 0, one, ctx);
    do {
        // Pas iteratiu.
        // Obtenir r_i(x).
        fq_poly_divrem(h, r_prev, r_prev, r, ctx);
        fq_poly_swap(r_prev, r, ctx);
        // Obtenir b_i(x).
        fq_poly_mul(h, h, b, ctx);
        fq_poly_sub(b_prev, b_prev, h, ctx);
        fq_poly_swap(b_prev, b, ctx);
    } while (fq_poly_degree(r, ctx) >= t);

    // Obtenir sigma.
    fq_poly_set(sigma, b, ctx);

    // Netejar la memoria.
    fq_poly_clear(h, ctx);
    fq_poly_clear(b_prev, ctx);
    fq_poly_clear(b, ctx);
    fq_poly_clear(r_prev, ctx);
    fq_poly_clear(r, ctx);

    if (fq_poly_degree(sigma, ctx) > t) {
        flint_printf("No s'ha pogut corregir l'error, ha fallat al pas 2.\n");
        return 1;
    }
    if (print) {
        flint_printf("Sigma:\t\t\t");
        fq_poly_print_pretty(sigma, "X", ctx);
        flint_printf("\n");
    }
    return 0;
}

slong decode(DecodingAlgorithm algorithm, slong index) {
    // Pas 1: Computar les sindromes.
    fq_t syndromes[d];
    slong error = 0;
    for (slong i = 1; i < d; i++) {
        fq_init(syndromes[i], ctx);
        fq_poly_evaluate_fq(syndromes[i], y[index], alpha_pows[i], ctx);
        if (!fq_is_zero(syndromes[i], ctx)) {
            error = 1;
        }
    }
    if (!error) {
        // Netegar la memoria.
        for (slong i = 1; i < d; i++) {
            fq_clear(syndromes[i], ctx);
        }
        if (print) flint_printf("No s'ha detectat errors.\n");
        return 0;
    }
    fq_poly_zero(sigma, ctx);
    // Pas 2:
    slong failed;
    switch (algorithm) {
        case Berlekamp_Massey:
            failed = decode_BM(syndromes);
            break;
        case Sugiyama:
            failed = decode_Sugiyama(syndromes);
            break;
        case Peterson_Gorenstein_Zierler:
        default:
            failed = decode_PGZ(syndromes);
            break;
    }
    // Netegar la memoria.
    for (slong i = 1; i < d; i++) {
        fq_clear(syndromes[i], ctx);
    }
    if (failed) {
        return 1;
    }

    // Pas 3: Obtenir els nombres de localitzacio d'error.
    slong nu = fq_poly_degree(sigma, ctx);
    slong loc[nu];
    slong temp = 0;
    for (slong i = 0; i < n && temp < nu; i++) {
        fq_poly_evaluate_fq(result, sigma, alpha_pows[i], ctx);
        if (fq_is_zero(result, ctx)) {
            if (i == 0)
                loc[temp++] = 0;
            else
                loc[temp++] = n - i;
        }
    }
    if (temp != nu) {
        if (print)
            flint_printf(
                "No s'ha pogut corregir l'error, ha fallat al pas 3.\n");
        return 1;
    }

    // Pas 4: Les magnituds d'error son 1 per ser un codi binari.
    fq_poly_zero(e, ctx);
    for (slong i = 0; i < nu; i++) {
        fq_poly_set_coeff(e, loc[i], one, ctx);
    }

    if (print) {
        flint_printf("Error detectat:\t\t");
        fq_poly_print_pretty(e, "X", ctx);
        flint_printf("\n");
    }

    // Obtenir el vector corregit, c_hat(x) = y(x) - e(x).
    fq_poly_sub(c_hat, y[index], e, ctx);
    if (print) {
        flint_printf("Codi corregit:\t\t");
        fq_poly_print_pretty(c_hat, "X", ctx);
        flint_printf("\n");
    }

    // Comprovar que el vector corregit es del codi.
    fq_poly_rem(r, c_hat, g, ctx);
    if (!fq_poly_is_zero(r, ctx)) {
        if (print)
            flint_printf(
                "No s'ha pogut corregir l'error, ha fallat a la "
                "comprovacio.\n");
        return 1;
    }

    // Obtenir el missatge original.
    fq_poly_shift_right(f_hat, c_hat, n - k, ctx);
    if (print) {
        flint_printf("Missatge corregit:\t");
        fq_poly_print_pretty(f_hat, "X", ctx);
        flint_printf("\n");
    }

    if (fq_poly_equal(f[index], f_hat, ctx)) {
        if (print) flint_printf("S'ha descodificat correctament.\n");
        return 0;
    } else {
        if (print) flint_printf("No s'ha descodificat correctament.\n");
        return 1;
    }
}

void generate_message() {
    for (slong i = 0; i < k; i++) {
        message[i] = rand() % 2;
    }
}

void generate_error() {
    for (slong i = 0; i < n; i++) {
        error_locations[i] = i;
    }
    // Remenar amb Fisher-Yates.
    for (slong i = n - 1; i > 0; i--) {
        slong j = rand() % (i + 1);
        slong temp = error_locations[i];
        error_locations[i] = error_locations[j];
        error_locations[j] = temp;
    }
}

slong geometric_random(slong max) {
    double random = rand() / (double)RAND_MAX;
    slong value = (slong)(log(random) / log(0.5));
    if (value > max) {
        return max;
    }
    return value;
}

void benchmark(slong iterations, slong max_errors, slong total_time) {
    clock_t start, end;
    slong time_PGZ = 0, time_BM = 0, time_S = 0;
    srand(time(NULL));
    num_errors = t;
    num_messages = iterations;
    init();
    for (slong i = 0; i < iterations; i++) {
        generate_message();
        encode(i);
        if (!max_errors) num_errors = geometric_random(255);
        generate_error();
        introduce_error(i);
    }
    // Peterson-Gorenstein-Zierler
    start = clock();
    for (slong i = 0; i < iterations; i++) {
        decode(Peterson_Gorenstein_Zierler, i);
        if (i % 1000 == 0) {
            flint_printf("%wd\n", i);
        }
    }
    end = clock();
    time_PGZ = end - start;
    // Berlekamp-Massey
    start = clock();
    for (slong i = 0; i < iterations; i++) {
        decode(Berlekamp_Massey, i);
        if (i % 1000 == 0) {
            flint_printf("%wd\n", i);
        }
    }
    end = clock();
    time_BM = end - start;
    // Sugiyama
    start = clock();
    for (slong i = 0; i < iterations; i++) {
        decode(Sugiyama, i);
        if (i % 1000 == 0) {
            flint_printf("%wd\n", i);
        }
    }
    end = clock();
    time_S = end - start;
    clear();
    if (total_time) {
        flint_printf("Temps total de descodificacio t=%wd:\n", t);
        flint_printf("PGZ: %.1f s\n",
                     time_PGZ / (double)(CLOCKS_PER_SEC));
        flint_printf("BM: %.1f s\n",
                     time_BM / (double)(CLOCKS_PER_SEC));
        flint_printf("S: %.1f s\n",
                     time_S / (double)(CLOCKS_PER_SEC));
    } else {
        flint_printf("Temps mitja de descodificacio t=%wd:\n", t);
        flint_printf("PGZ: %.3f ms\n",
                    time_PGZ / (double)((CLOCKS_PER_SEC / 1000) * iterations));
        flint_printf("BM: %.3f ms\n",
                    time_BM / (double)((CLOCKS_PER_SEC / 1000) * iterations));
        flint_printf("S: %.3f ms\n",
                    time_S / (double)((CLOCKS_PER_SEC / 1000) * iterations));
    }
}

// Compilar: gcc main.c -lflint -lm -o bch
// Executar: ./bch [benchmark] <mode>
int main(int argc, char *argv[]) {
    if ((argc == 2 || argc == 3) &&
        (argv[1] && strcmp(argv[1], "benchmark") == 0)) {
        slong mode = 0;
        if (argc == 3) mode = atoi(argv[2]);
        switch (mode) {
            case 1:
                flint_printf("Comparacio mode 1\n");
                m = 8;
                t = 5;
                benchmark(2048, 1, 0);
                t = 10;
                benchmark(2048, 1, 0);
                t = 15;
                benchmark(2048, 1, 0);
                t = 20;
                benchmark(2048, 1, 0);
                t = 25;
                benchmark(2048, 1, 0);
                return 0;
            case 2:
                flint_printf("Comparacio mode 2\n");
                m = 6;
                t = 2;
                benchmark(2048, 1, 0);
                t = 4;
                benchmark(2048, 1, 0);
                t = 6;
                benchmark(2048, 1, 0);
                t = 10;
                benchmark(2048, 1, 0);
                t = 15;
                benchmark(2048, 1, 0);
                return 0;
            case 3:
                flint_printf("Comparacio mode 3\n");
                m = 8;
                t = 10;
                benchmark(134079, 0, 1);
                return 0;
            case 0:
            default:
                flint_printf("Comparacio mode 0\n");
                m = 8;
                t = 5;
                benchmark(2048, 0, 0);
                t = 10;
                benchmark(2048, 0, 0);
                t = 15;
                benchmark(65536, 0, 0);
                t = 20;
                benchmark(65536, 0, 0);
                t = 25;
                benchmark(65536, 0, 0);
                return 0;
        }
    }
    print = 1;
    flint_printf("Introdueix m, on 2^m-1 es la longitud del codi:\n");
    flint_scanf("%wd", &m);
    if (m < 2 || m > 8) {
        flint_printf("m ha de ser un enter entre 2 i 8.\n");
        return 1;
    }
    flint_printf("Introdueix t, el nombre de bits d'error corregibles:\n");
    flint_scanf("%wd", &t);
    if (t < 1 || t > n_pow(2, m) - 1) {
        flint_printf("t ha de ser un enter entre 1 i 2^m-1.\n");
        return 1;
    }
    srandom(time(NULL));
    init();
    flint_printf("Introdueix el nombre de bits d'error a introduir:\n");
    flint_scanf("%wd", &num_errors);
    if (num_errors < 0 || num_errors > n_pow(2, m) - 1) {
        flint_printf(
            "El nombre d'errors ha de ser un enter entre 0 i 2^m-1.\n");
        return 1;
    }
    generate_message();
    encode(0);
    generate_error();
    introduce_error(0);
    flint_printf("Per Peterson_Gorenstein_Zierler:\n");
    decode(Peterson_Gorenstein_Zierler, 0);
    flint_printf("Per Berlekamp_Massey:\n");
    decode(Berlekamp_Massey, 0);
    flint_printf("Per Sugiyama:\n");
    decode(Sugiyama, 0);
    clear();
}